
#ifndef LCD_TEXT_H
#define LCD_TEXT_H

//============================================================================//

// Port:
#define LCD_DDR  DDRD
#define LCD_PORT PORTD
#define LCD_PIN  PIND

// Pin-Belegung:

/* 2x16:
 * RS = 4
 * RW = 8
 * E1 = 5
 * E2 = 8
 * D4 = 0
 * D5 = 1
 * D6 = 2
 * D7 = 3
 */
#define LCD_PIN_RS PD0
#define LCD_PIN_RW PD1
#define LCD_PIN_E1 PD2
#define LCD_PIN_E2 PD3
#define LCD_PIN_D4 PD4
#define LCD_PIN_D5 PD5
#define LCD_PIN_D6 PD6
#define LCD_PIN_D7 PD7

// Wenn die frequenz noch nicht gesetzt wurde hier setzten.
#ifndef F_CPU
	#define F_CPU 8000000UL
#endif

//============================================================================//

#if LCD_PIN_RS > 7
	#define LCD_RS 0
#else
	#define LCD_RS (1 << LCD_PIN_RS)
#endif

#if LCD_PIN_RW > 7
	#define LCD_RW 0
#else
	#define LCD_RW (1 << LCD_PIN_RW)
#endif

#if LCD_PIN_E1 > 7
	#define LCD_E1 0
#else
	#define LCD_E1 (1 << LCD_PIN_E1)
#endif

#if LCD_PIN_E2 > 7
	#define LCD_E2 0
#else
	#define LCD_E2 (1 << LCD_PIN_E2)
#endif

#if LCD_PIN_D4 > 7
	#define LCD_D4 0
#else
	#define LCD_D4 (1 << LCD_PIN_D4)
#endif

#if LCD_PIN_D5 > 7
	#define LCD_D5 0
#else
	#define LCD_D5 (1 << LCD_PIN_D5)
#endif

#if LCD_PIN_D6 > 7
	#define LCD_D6 0
#else
	#define LCD_D6 (1 << LCD_PIN_D6)
#endif

#if LCD_PIN_D7 > 7
	#define LCD_D7 0
#else
	#define LCD_D7 (1 << LCD_PIN_D7)
#endif

//============================================================================//

#include <inttypes.h>

#include <util/delay.h>
#include <avr/io.h>

//============================================================================//

static uint8_t lcd_shift_port(uint8_t data, uint8_t rs, uint8_t rw, uint8_t enable);

static void lcd_write(uint8_t data, uint8_t rs, uint8_t enable);

void lcd_write_control(uint8_t command, uint8_t enable);
void lcd_write_data   (uint8_t data, uint8_t enable);

void lcd_init(void);
void lcd_clear(void);
void lcd_position(uint8_t zeile, uint8_t spalte, uint8_t enable);
void lcd_text(const char *text, uint8_t length, uint8_t enable);

//============================================================================//

static uint8_t lcd_shift_port(uint8_t data, uint8_t rs, uint8_t rw, uint8_t enable)
{
	uint8_t port = 0x00;
	
	data = data & 0x0F;
	
	if((rs & ~LCD_RS) == 0)
		port |= rs;
	
	if((rw & ~LCD_RW) == 0)
		port |= rw;
	
	if((enable & ~LCD_E1 & ~LCD_E2) == 0)
		port |= enable;

	if((data & 0x01) != 0)
		port |= LCD_D4;
	
	if((data & 0x02) != 0)
		port |= LCD_D5;
	
	if((data & 0x04) != 0)
		port |= LCD_D6;
	
	if((data & 0x08) != 0)
		port |= LCD_D7;
	
	return port;
}

//============================================================================//

static void lcd_write(uint8_t data, uint8_t rs, uint8_t enable)
{
	// Datenbit 4 bis 7 bertragen.	
	
	// Ausmaskieren der niederwertigen bits. Signal wird auf den port gelegt.
	LCD_PORT = lcd_shift_port((data & 0xF0) >> 4, rs, 0, 0);
	
	// Zum besttigen der daten E = 1.
	LCD_PORT = lcd_shift_port((data & 0xF0) >> 4, rs, 0, enable);

	// Mindestens 220ns warten
	asm("nop");
	asm("nop");
	asm("nop");
	
	// Enable wieder auf 0 setzten.
	LCD_PORT = LCD_PORT & ~enable;
	
	// Datenbit 0 bis 3 bertragen.
	
	// Ausmaskieren des hherwertigen bits. Signal wird auf dem port gelegt.
	LCD_PORT = lcd_shift_port((data & 0x0F), rs, 0, 0);
	    							         	
	// Zum besttigen der daten E = 1.
	LCD_PORT = lcd_shift_port((data & 0x0F), rs, 0, enable);

	// Mindestens 220ns warten.
	asm("nop");
	asm("nop");
	asm("nop");
	// Enable wieder auf 0 setzten.
	LCD_PORT = LCD_PORT & ~enable;
	
	// Mindestens 42 us warten.
	_delay_us(50);
}

//============================================================================//

void lcd_write_control(uint8_t command, uint8_t enable)
{
	lcd_write(command, 0, enable);
}

//============================================================================//

void lcd_write_data(uint8_t data, uint8_t enable)
{
	lcd_write(data, LCD_RS, enable);
}

//============================================================================//

void lcd_init(void)
{
	LCD_DDR = 0xff;

	// Mindestens 17ms warten, nachdem die Betriebsspannung angelegt wurde.
	_delay_ms(17);
	// System set: vierzeiliges Display, 4 - Bit - Modus
	lcd_write_control(0x28, LCD_E1 | LCD_E2);
	
	// Mindestens 4.1ms warten.
	_delay_ms(5);
	lcd_write_control(0x28, LCD_E1 | LCD_E2);
	
	// Mindestens 100us warten.
	_delay_us(120);
	lcd_write_control(0x28, LCD_E1 | LCD_E2);
	
	// Warten bis alle arbeiten erledigt sind. Eigentlich 
	// msste das busy flag abgefragt werden.
	_delay_us(100);

	// Display wird eingeschaltet, Cursor wird ausgeschaltet
	lcd_write_control(0x0C, LCD_E1 | LCD_E2);

	// Schreibrichtung wird festgelegt.
	lcd_write_control(0x06, LCD_E1 | LCD_E2);
	
	// Platzierung des Cursors an Position 0.
	lcd_write_control(0x02, LCD_E1 | LCD_E2);           		
	_delay_ms (2);
	
	// Display einmal lschen.
	lcd_clear();
}

//============================================================================//

void lcd_clear(void)
{
	// Diesplay leeren.
	lcd_write_control(0x01, LCD_E1 | LCD_E2);

	// Midestens 1.64ms warten.
	_delay_ms (2);
}

//============================================================================//

void lcd_position(uint8_t zeile, uint8_t spalte, uint8_t enable)
{
	lcd_write_control(0x80 + (spalte - 1) + 0x40*(zeile - 1), enable);
}

//============================================================================//

void lcd_text(const char *text, uint8_t length, uint8_t enable)
{
	uint8_t i;
	
	for(i = 0; i < length; i++)
	{
		lcd_write_data(text[i], enable);
	}
}

//============================================================================//

#endif // LCD_TEXT_H



